#include "libzip.h"

#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef __WIN32__
#include <windows.h>
#else
#include <libgen.h>
#endif

#ifdef __WIN32__

static int mkdir(const char *path, int mode)
{
    return (CreateDirectory(path, NULL) ? 0 : -1);
}

static char *dirname(char *path) {
    char *s1 = strrchr(path, '/');
    char *s2 = strrchr(path, '\\');

    if (s1 && s1 >= s2) {
        *s1 = '\0';
    } else if (s2) {
        *s2 = '\0';
    } else {
        *path = '\0';
    }

    return path;
}

#endif

static void print_usage()
{
    printf(
        "Usage: dezip [OPTIONS] [FILENAME]\n"
        "Decompresses the PKZIP archive specified by filename. If no filename is passed\n"
        "the archive is read from standard input.\n"
        "\n"
        "Options:\n"
        "  -d BITMASK    debug messages to show\n"
        "  -p PASSWORD   password to use\n"
        "  -s            write content to stdout instead of files\n"
        "  -t            just test, do not extract\n"
        "\n"
        "Copyright 2005 Mathias Hasselmann <mathias@taschenorakel.de>\n"
        "Bug reports: https://taschenorakel.de/bugzilla/enter_bug.cgi?product=libzip\n");
}

static int pstrcmp(const void *a, const void *b)
{
    return strcmp(*(char**)a, *(char**)b);
}

int main(int argc, char* argv[]) {
    const char *password = NULL;
    FILE *istream = NULL;
    bool extract = true;
    bool pipe = false;
    int verbose = 1;
    int i;

    libzip_stream_t *stream;
    libzip_reader_t *reader;
    libzip_entry_t *entry;

    while(-1 != (i = getopt(argc, argv, "d:p:sth"))) {
        switch(i) {
            case 'd':
                libzip_set_debug_level(atoi(optarg));
                verbose = 2;
                break;

            case 'p':
                password = optarg;
                break;
                
            case 's':
                pipe = true;
                break;

            case 't':
                extract = false;
                break;

            case 'h':
                print_usage();
                return 0;

            default:
                fprintf(stderr, "Usage error. Call 'dezip -h' for help.\n");
                return 1;
        }
    }

    if (optind == argc || !strcmp(argv[optind], "-")) {
        istream = stdin;
    } else if (NULL == (istream = fopen(argv[optind++], "rb"))) {
        perror(argv[optind - 1]);
        return 1;
    }

    if (pipe)
        verbose = 0;

    char **filters = argv + optind;
    const int flen = argc - optind;

    qsort(filters, flen, sizeof(char*), pstrcmp);

    stream = libzip_stdio_stream_open(istream ? istream : stdin);
    reader = libzip_reader_open(stream);

    if (password)
        libzip_reader_set_password(reader, password);

    for(entry = libzip_reader_find_entry(reader, NULL); entry;
        entry = libzip_reader_next_entry(reader, entry)) {
        const char *fname = libzip_entry_get_filename(entry);
        char *dcopy = strdup(fname);
        char *dname = dirname(dcopy);

        if (verbose > 0)
            printf("%s (%d)\n", fname, libzip_entry_get_length(entry));

        bool skip = 
            optind < argc && NULL ==
            bsearch(&fname, filters, flen, sizeof(char*), pstrcmp);

        if (skip && verbose > 0)
            printf("! skipping %s\n", fname);

        FILE *ostream = NULL;

        if (extract && !skip) {
            if (*dname)
                mkdir(dname, 0777);

            if (pipe) {
                ostream = stdout;
            } else if (NULL == (ostream = fopen(fname, "w"))) {
                perror(fname);
            }
        }

        char buffer[65536];
        size_t remaining = 0;
        size_t length = 0;

        while(0 != (remaining = libzip_reader_get_remaining(reader)) &&
              0 != (length = libzip_reader_read_current(reader, buffer, sizeof(buffer)))) {
            if (verbose > 1)
                printf("| remaining: %d, length: %d\n", remaining, length);

            if (ostream && length != fwrite(buffer, 1, length, ostream))
                perror("Cannot write file");
        }

        if (verbose > 1)
            printf("\\ remaining: %d, length: %d\n", remaining, length);

        if (ostream && ostream != stdout)
            fclose(ostream);
        if (dcopy)
            free(dcopy);
    }

    libzip_reader_close(reader);
    libzip_stream_close(stream);

    if (istream && stdin != istream)
        fclose(istream);

    return 0;
}

/* vim: set sw=4 sta et: */
