/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/**
 * \file libzip-datetime.c
 * Date and time related functions.
 *
 * As the ZIP file format has its origins on MS-DOS it stores timestamps 
 * in MS-DOS FAT format. This format stores dates starting from 1980/01/01
 * with a resolution of two seconds.
 */

#include "libzip-private.h"

#include <string.h>

/**
 * Converts a POSIX timestamp into a MS-DOS timestamp.
 *
 * @param timestamp The POSIX timestamp to convert.
 * @return The aquivalent MS-DOS timestamp.
 */
uint32_t libzip_dos_date_time(const time_t *timestamp)
{
    uint32_t dos_date, dos_time;
    struct tm *posix;

    return_val_if_fail(NULL != timestamp, 0);
    posix = localtime(timestamp);

    dos_date = ((posix->tm_year - 80) << 9) + ((posix->tm_mon + 1) << 5) + posix->tm_mday;
    dos_time = (posix->tm_hour << 11) + (posix->tm_min << 5) + (posix->tm_sec >> 1);

    return ((dos_date << 16) | dos_time);
}

/**
 * Converts a MS-DOS timestamp into a POSIX timestamp.
 *
 * @param timestamp The MS-DOS timestamp to convert.
 * @return The aquivalent POSIX timestamp.
 */
time_t libzip_timestamp(const uint32_t timestamp)
{
    struct tm posix;

    const uint32_t dos_date = (timestamp >> 16);
    const uint32_t dos_time =  timestamp;

    memset(&posix, 0, sizeof(posix));

    posix.tm_year = ((dos_date >> 9) & 0x7f) + 80;
    posix.tm_mon  = ((dos_date >> 5) & 0x0f) - 1;
    posix.tm_mday =  (dos_date       & 0x1f);

    posix.tm_hour = ((dos_time >> 11) & 0x1f);
    posix.tm_min  = ((dos_time >>  5) & 0x3f);
    posix.tm_sec  = ((dos_time <<  1) & 0x3f);

    return mktime(&posix);
}

/* vim: set sw=4 sta et: */
