#ifndef LIBZIP_PRIVATE_INCLUDED
#define LIBZIP_PRIVATE_INCLUDED

#include "libzip.h"
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ===== general purpose macros ===== */

#define libzip_new(Type, Count) (Type*) calloc((Count), sizeof(Type))
#define LIBZIP_SPECIFICATION_VERSION(Major, Minor) ((Major) * 10 + (Minor))

/* ===== error handling and debugging ===== */

#ifndef __cplusplus

#ifdef __WIN32
#define bold(Str)       Str
#else
#define bold(Str)       "\x1b[1m" Str "\x1b[0m"
#endif

//#define TRACING_ENABLED
#ifdef TRACING_ENABLED

#define trace(Domain, Message) \
    fprintf(stderr, "[libzip] %s (%s, line %d)\n" \
                    "[libzip] " bold(Domain ": " Message) "\n", \
                    __FUNCTION__, __FILE__, __LINE__)

#define tracef(Domain, Message, ...) \
    fprintf(stderr, "[libzip] %s (%s, line %d)\n" \
                    "[libzip] " bold(Domain ": " Message) "\n", \
                    __FUNCTION__, __FILE__, __LINE__, __VA_ARGS__)

#else

#define trace(Domain, Message) do {} while(0)
#define tracef(Domain, Message, ...) do {} while(0)

#endif // TRACING_ENABLED

#define error(Message) trace("ERROR", Message)
#define errorf(Message, ...) tracef("ERROR", Message, __VA_ARGS__)
                                                              
#define warn(Message) trace("WARNING", Message)
#define warnf(Message, ...) trace("WARNING", Message, __VA_ARGS__)

#define raise(Message) do { error(Message " Aborting."); goto finally; } while(0)
#define raisef(Message, ...) do { errorf(Message " Aborting.", __VA_ARGS__); goto finally; } while(0)

#define debug(Message) do { \
    if (__builtin_expect(libzip_debug_level & 1, 0)) trace("DEBUG", Message); \
} while(0)

#define debugf(Message, ...) do { \
    if (__builtin_expect(libzip_debug_level & 1, 0)) tracef("DEBUG", Message, __VA_ARGS__); \
} while(0)

#define raise_if_fail(Assertion) do { \
    if (__builtin_expect(!(Assertion), 0)) raise("Assertion failed: " #Assertion "."); \
} while(0)

#define warn_if_fail(Assertion) do { \
    if (__builtin_expect(!(Assertion), 0)) warn("Assertion failed: " #Assertion "."); \
} while(0)

#define return_val_if_fail(Assertion, Value) do { \
    if (__builtin_expect(!(Assertion), 0)) { \
        error("Assertion failed: " #Assertion ". Returning " #Value "."); \
        return (Value); \
    } \
} while(0)

#define return_if_fail(Assertion) do { \
    if (__builtin_expect(!(Assertion), 0)) { \
        error("Assertion failed: " #Assertion ". Returning."); \
        return; \
    } \
} while(0)

#endif

/* ===== symbol attributes ===== */

#define LIBZIP_PACKED          __attribute__((packed))
                                                              
#ifndef __WIN32
#define LIBZIP_INTERNAL        __attribute__((visibility("internal")))
#else
#define LIBZIP_INTERNAL
#endif

/* ===== enumerations and constants ===== */

enum {
    LIBZIP_SIGNATURE_LOCAL_HEADER =             0x04034b50,
    LIBZIP_SIGNATURE_DIRECTORY_RECORD =         0x02014b50,
    LIBZIP_SIGNATURE_END_OF_DIRECTORY =         0x06054b50,
    LIBZIP_SIGNATURE_LOCAL_DESCRIPTOR =         0x08074b50
};

enum {
    LIBZIP_EXTRA_SIGNATURE_WINZIP_AES =         0x9901
};

enum {
    LIBZIP_WINZIP_AES_VENDOR_AE =               0x4541
};

enum {
    LIBZIP_WINZIP_AES_VERSION_AE1 =             0x0001,
    LIBZIP_WINZIP_AES_VERSION_AE2 =             0x0002
};

typedef enum {
    LIBZIP_FEATURE_LEVEL_DEFAULT =              10,
    LIBZIP_FEATURE_LEVEL_VOLUME_LABEL =         11,
    LIBZIP_FEATURE_LEVEL_FOLDER =               20,
    LIBZIP_FEATURE_LEVEL_DEFLATE =              20,
    LIBZIP_FEATURE_LEVEL_ENCRYPTION =           20,
    LIBZIP_FEATURE_LEVEL_DEFLATE64 =            21,
    LIBZIP_FEATURE_LEVEL_IMPLODE =              25,
    LIBZIP_FEATURE_LEVEL_PATCH_DATA_SET =       27,
    LIBZIP_FEATURE_LEVEL_ZIP64 =                45,
    LIBZIP_FEATURE_LEVEL_BZIP2 =                46,
    LIBZIP_FEATURE_LEVEL_DES =                  50,
    LIBZIP_FEATURE_LEVEL_3DES =                 50,
    LIBZIP_FEATURE_LEVEL_RC2_BROKEN =           50,
    LIBZIP_FEATURE_LEVEL_RC4 =                  50,
    LIBZIP_FEATURE_LEVEL_AES =                  51,
    LIBZIP_FEATURE_LEVEL_RC2 =                  51,
    LIBZIP_FEATURE_LEVEL_RC2_64 =               52,
    LIBZIP_FEATURE_LEVEL_KEY_WRAPPING =         61,
    LIBZIP_FEATURE_LEVEL_DIRECTORY_ENCRPTION =  62
} libzip_feature_level_t;

typedef enum {
    LIBZIP_FILESYSTEM_MSDOS, // and OS/2 (FAT / VFAT / FAT32 file systems)
    LIBZIP_FILESYSTEM_AMIGA,
    LIBZIP_FILESYSTEM_OPENVMS,
    LIBZIP_FILESYSTEM_UNIX,
    LIBZIP_FILESYSTEM_VM_CMS,
    LIBZIP_FILESYSTEM_ATARI_ST,
    LIBZIP_FILESYSTEM_OS2_HPFS,
    LIBZIP_FILESYSTEM_MACINTOSH,
    LIBZIP_FILESYSTEM_Z_SYSTEM,
    LIBZIP_FILESYSTEM_CPM,
    LIBZIP_FILESYSTEM_WINDOWS_NTFS,
    LIBZIP_FILESYSTEM_MVS, // (OS/390 - Z/OS)
    LIBZIP_FILESYSTEM_VSE,
    LIBZIP_FILESYSTEM_ACORN_RISC,
    LIBZIP_FILESYSTEM_VFAT,
    LIBZIP_FILESYSTEM_ALTERNATE_MVS,
    LIBZIP_FILESYSTEM_BEOS,
    LIBZIP_FILESYSTEM_TANDEM,
    LIBZIP_FILESYSTEM_OS400,
    LIBZIP_FILESYSTEM_OSX
} libzip_writersystem_t;

typedef enum {
    LIBZIP_FLAG_ENCRYPTED =             (1 << 0),

    LIBZIP_FLAG_IMPLODE_8K =            (1 << 1),
    LIBZIP_FLAG_IMPLODE_SHANNON_FANO =  (1 << 2),

    LIBZIP_FLAG_DEFLATE_NORMAL =        (0 << 1),
    LIBZIP_FLAG_DEFLATE_MAXIMUM =       (1 << 1),
    LIBZIP_FLAG_DEFLATE_FAST =          (2 << 1),
    LIBZIP_FLAG_DEFLATE_SUPER_FAST =    (3 << 1),

    LIBZIP_FLAG_DESCRIPTOR_NEEDED =     (1 << 3),

    LIBZIP_FLAG_RESERVED_DEFLATE =      (1 << 4),
    LIBZIP_FLAG_COMPRESSED_PATCH_DATA = (1 << 5),
    LIBZIP_FLAG_STRONG_ENCRYPTION =     (1 << 6),

    LIBZIP_FLAG_PKWARE_RESERVED_1 =     (1 << 12),
    LIBZIP_FLAG_LOCAL_HEADER_MASKED  =  (1 << 13),
    LIBZIP_FLAG_PKWARE_RESERVED_2 =     (1 << 14),
    LIBZIP_FLAG_PKWARE_RESERVED_3 =     (1 << 15)
} libzip_flag_t;

typedef enum {
    LIBZIP_INTERNAL_ATTRIBUTE_ASCII =   (1 << 0)
} libzip_internal_attribute_t;

/* ===== zip file headers ===== */

typedef struct {
    uint32_t signature;         // LIBZIP_SIGNATURE_LOCAL_HEADER
    uint16_t feature_level;
    uint16_t flags;
    uint16_t compression;
    uint32_t last_modified;
    uint32_t checksum;
    uint32_t compressed_size;
    uint32_t uncompressed_size;
    uint16_t filename_length;
    uint16_t extra_length;
} LIBZIP_PACKED libzip_local_header_t;

typedef struct {
    /**
     * Signature of this file header: LIBZIP_SIGNATURE_LOCAL_DESCRIPTOR.
     * Oddly enough <tt>appnote.txt</tt> misses to mention it.
     */
    uint32_t signature;
    uint32_t checksum;
    uint32_t compressed_size;
    uint32_t uncompressed_size;
} LIBZIP_PACKED libzip_local_descriptor_t;

typedef struct {
    uint16_t signature;
    uint16_t length;
} LIBZIP_PACKED libzip_extra_t;

typedef struct {
    libzip_extra_t header;
    uint16_t version;
    uint16_t vendor;
    uint8_t mode;
    uint16_t compression;
} LIBZIP_PACKED libzip_winzip_aes_t;

typedef struct {
    uint32_t signature;         // LIBZIP_SIGNATURE_DIRECTORY_RECORD
    uint8_t creator_version;
    uint8_t filesystem;
    uint16_t feature_level;
    uint16_t flags;
    uint16_t compression;
    uint32_t last_modified;
    uint32_t checksum;
    uint32_t compressed_size;
    uint32_t uncompressed_size;
    uint16_t filename_length;
    uint16_t extra_length;
    uint16_t comment_length;
    uint16_t disk_number_start;
    uint16_t internal_attributes;
    uint32_t external_attributes;
    uint32_t offset_of_local_header;
} LIBZIP_PACKED libzip_directory_record_t;

typedef struct {
    uint32_t signature;         // LIBZIP_SIGNATURE_END_OF_DIRECTORY
    uint16_t disk_number;
    uint16_t central_directory_disk;
    uint16_t entries_on_disk_count;
    uint16_t total_entry_count;
    uint32_t length;
    uint32_t offset;
    uint16_t comment_length;
} LIBZIP_PACKED libzip_end_of_directory_t;

/* ===== LIBZIP_INTERNAL structures ===== */

typedef struct libzip_decoder_t libzip_decoder_t;
typedef struct libzip_encoder_t libzip_encoder_t;

struct libzip_entry_t {
    libzip_entry_t *next;

    size_t offset;
    char *filename;
    uint8_t *extra;

    uint16_t internal_attributes;
    libzip_local_header_t header;
};

/* ===== libzip ===== */

void libzip_random(uint8_t *buffer, size_t length) LIBZIP_INTERNAL;

/* ===== libzip-datetime ===== */

uint32_t libzip_dos_date_time(const time_t *timestamp) LIBZIP_INTERNAL;
time_t libzip_timestamp(const uint32_t dos_date_time) LIBZIP_INTERNAL;

/* ===== libzip-decoder ===== */

libzip_decoder_t* libzip_decoder_new(libzip_stream_t *stream, libzip_entry_t *entry, const char *password) LIBZIP_INTERNAL;
size_t libzip_decoder_read(libzip_decoder_t *decoder, void *data, size_t length) LIBZIP_INTERNAL;
void libzip_decoder_free(libzip_decoder_t* decoder) LIBZIP_INTERNAL;

#if ENABLE_PPMD
int libzip_ppmd_decoder_getc(libzip_decoder_t *decoder) LIBZIP_INTERNAL;
int libzip_ppmd_decoder_putc(libzip_decoder_t *decoder, int c) LIBZIP_INTERNAL;
#endif

/* ===== libzip-encoder ===== */

libzip_encoder_t* libzip_encoder_new(libzip_stream_t *stream, libzip_entry_t *entry, const char *password) LIBZIP_INTERNAL;
bool libzip_encoder_write(libzip_encoder_t *encoder, const void *data, size_t length) LIBZIP_INTERNAL;
bool libzip_encoder_stat(libzip_encoder_t *encoder, uint32_t *uncompressed_size,
                         uint32_t *compressed_size, uint32_t *checksum) LIBZIP_INTERNAL;
void libzip_encoder_free(libzip_encoder_t* encoder) LIBZIP_INTERNAL;

#if ENABLE_PPMD
int libzip_ppmd_encoder_getc(libzip_encoder_t *encoder) LIBZIP_INTERNAL;
int libzip_ppmd_encoder_putc(libzip_encoder_t *encoder, int c) LIBZIP_INTERNAL;
#endif

/* ===== libzip-entry ===== */

libzip_entry_t* libzip_entry_new(const char *filename, const time_t *timestamp) LIBZIP_INTERNAL;
void libzip_entry_add_extra(libzip_entry_t *entry, const libzip_extra_t *extra) LIBZIP_INTERNAL;
const libzip_extra_t* libzip_entry_find_extra(const libzip_entry_t *entry, uint16_t signature);
void libzip_entry_free(libzip_entry_t *entry) LIBZIP_INTERNAL;

/* ===== libzip-stream ===== */

libzip_stream_t* libzip_aes_stream_open(libzip_stream_t* wrapped, libzip_aes_mode_t mode, const char *password) LIBZIP_INTERNAL;

long libzip_stream_tell(libzip_stream_t *stream) LIBZIP_INTERNAL;
bool libzip_stream_seek(libzip_stream_t *stream, long offset, int whence) LIBZIP_INTERNAL;
size_t libzip_stream_read(libzip_stream_t *stream, void *data, size_t length) LIBZIP_INTERNAL;
size_t libzip_stream_write(libzip_stream_t *stream, const void *data, size_t length) LIBZIP_INTERNAL;
bool libzip_stream_flush(libzip_stream_t *stream) LIBZIP_INTERNAL;

/* ===== libzip-debug ===== */

extern int libzip_debug_level LIBZIP_INTERNAL;

void libzip_debug_print_local_header(libzip_local_header_t *header, const char *filename, uint8_t *extra);
void libzip_debug_print_directory_record(libzip_directory_record_t *record, const char *filename, uint8_t *extra);
void libzip_debug_print_end_of_directory(libzip_end_of_directory_t *end);
void libzip_debug_print_local_descriptor(libzip_local_descriptor_t *descriptor);

#ifdef __cplusplus
}
#endif

#endif /* vim: set sw=4 sta et: */
