/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "libzip-private.h"

#include <stdlib.h>
#include <string.h>

/* ===== private structures ===== */

struct libzip_reader_t {
    libzip_stream_t *stream;
    libzip_decoder_t *decoder;

    libzip_entry_t *current;
    libzip_entry_t *entries;
    unsigned entry_count;
    size_t remaining;
    
    char *password;
};

/* ===== libzip_reader ===== */

static bool libzip_reader_read_central_directory(libzip_reader_t *reader)
{
    bool success = false;

    libzip_directory_record_t record;
    libzip_end_of_directory_t end;

    size_t bufsize = 0;
    char *filename = NULL;
    time_t timestamp;

    if (false == libzip_stream_seek(reader->stream, -sizeof(end), SEEK_END))
        raise("Cannot seek to end of central directory record.");
    if (sizeof(end) != libzip_stream_read(reader->stream, &end, sizeof(end)))
        raise("Cannot read end of central directory record.");

    libzip_debug_print_end_of_directory(&end);

    if (LIBZIP_SIGNATURE_END_OF_DIRECTORY != end.signature)
        raise("Bad signature at position of end of central directory record.");
    if (false == libzip_stream_seek(reader->stream, end.offset, SEEK_SET))
        raise("Cannot seek to first entry of central directory record.");

    while(reader->entry_count < end.entries_on_disk_count) {
        libzip_entry_t *entry = NULL;
        uint8_t *extra = NULL;

        if (sizeof(record) != libzip_stream_read(reader->stream, &record, sizeof(record)))
            raisef("Cannot read central directory record #%d.", reader->entry_count);

        if (LIBZIP_SIGNATURE_DIRECTORY_RECORD != record.signature)
            raisef("Bad signature at position of central directory record #%d.", reader->entry_count);

        if (NULL == filename || bufsize < record.filename_length) {
            if (filename)
                free(filename);
            
            bufsize = record.filename_length;
            filename = libzip_new(char, bufsize + 1);
        }

        if (record.filename_length != libzip_stream_read(reader->stream, filename, record.filename_length))
            raisef("Cannot read filename of central directory record #%d.", reader->entry_count);

        filename[record.filename_length] = '\0';
        timestamp = libzip_timestamp(record.last_modified);

        if (record.extra_length) {
            extra = libzip_new(uint8_t, record.extra_length);
            libzip_stream_read(reader->stream, extra, record.extra_length);
        }
        
        if (record.comment_length)
            libzip_stream_seek(reader->stream, record.comment_length, SEEK_CUR);

        libzip_debug_print_directory_record(&record, filename, extra);

        entry = libzip_entry_new(filename, &timestamp);
        return_val_if_fail(NULL != entry, false);

        entry->header.feature_level = record.feature_level;
        entry->header.flags = record.flags;
        entry->header.compression = record.compression;
        entry->header.last_modified = record.last_modified;
        entry->header.checksum = record.checksum;
        entry->header.compressed_size = record.compressed_size;
        entry->header.uncompressed_size = record.uncompressed_size;
        entry->header.extra_length = record.extra_length;

        entry->offset = record.offset_of_local_header;
        entry->extra = extra; extra = NULL;
        entry->next = reader->entries;


        reader->entries = entry;
        ++reader->entry_count;
    }

    success = true;

finally:
    if (filename)
        free(filename);

    return success;
}

libzip_reader_t* libzip_reader_open(libzip_stream_t *stream) {
    libzip_reader_t *reader;

    return_val_if_fail(NULL != stream, NULL);
    reader = libzip_new(libzip_reader_t, 1);

    reader->stream = stream;

    if (!libzip_reader_read_central_directory(reader)) {
        libzip_reader_close(reader);
        reader = NULL;
    }

    return reader;
}

void libzip_reader_set_password(libzip_reader_t *reader, const char* password)
{
    return_if_fail(NULL != reader);

    if (reader->password) {
        free(reader->password);
        reader->password = NULL;
    }

    if (password)
        reader->password = strdup(password);
}

bool libzip_reader_seek(libzip_reader_t *reader, const libzip_entry_t* entry) {
    bool success = false;
    libzip_local_header_t header;

    raise_if_fail(NULL != reader);
    raise_if_fail(NULL != reader->stream);
    raise_if_fail(NULL != entry);

    if (reader->decoder) {
        libzip_decoder_free(reader->decoder);
        reader->decoder = NULL;
    }
    
    reader->current = NULL;
    reader->decoder = libzip_decoder_new(reader->stream, entry, reader->password);
    raise_if_fail(NULL != reader->decoder);
    
    if (libzip_stream_seek(reader->stream, entry->offset, SEEK_SET) &&
        libzip_stream_read(reader->stream, &header, sizeof(header)) &&
        libzip_stream_seek(reader->stream, header.filename_length + header.extra_length, SEEK_CUR)) {
        reader->remaining = libzip_entry_get_length(entry);
        reader->current = entry;
        success = true;
    }

finally:
    return success;
}

libzip_entry_t* libzip_reader_find_entry(libzip_reader_t *reader, const char *filename, bool case_sensitive) {
    libzip_entry_t* entry;
    
    return_val_if_fail(NULL != reader, NULL);
    return_val_if_fail(NULL != reader->stream, NULL);

    entry = reader->entries;

    if (filename) 
    {
		if(case_sensitive)
		{
	        while(entry && strcmp(filename, entry->filename))
	            entry = entry->next;
		}
		else
		{
	        while(entry && strcasecmp(filename, entry->filename))
	            entry = entry->next;
		}
    }

    if (entry)
        libzip_reader_seek(reader, entry);

    return entry;
}

libzip_entry_t* libzip_reader_next_entry(libzip_reader_t *reader, libzip_entry_t* entry) {
    return_val_if_fail(NULL != reader, NULL);
    return_val_if_fail(NULL != reader->stream, NULL);
    return_val_if_fail(NULL != entry, NULL);

    if (NULL != (entry = entry->next))
        libzip_reader_seek(reader, entry);

    return entry;
}

void libzip_reader_close(libzip_reader_t *reader) {
    if (reader) {
        libzip_entry_t *entry;
        
        if (reader->decoder)
            libzip_decoder_free(reader->decoder);

        while(NULL != (entry = reader->entries)) {
            reader->entries = entry->next;
            libzip_entry_free(entry);
        }

        if (reader->password)
            free(reader->password);

        free(reader);
    }
}

size_t libzip_reader_read_current(libzip_reader_t *reader, void *buffer, size_t length)
{
    size_t bytes_read = 0;
    
    raise_if_fail(NULL != reader);

    if (length > reader->remaining)
        length = reader->remaining;
    
    bytes_read = libzip_decoder_read(reader->decoder, buffer, length);
    reader->remaining-= bytes_read;

finally:
    return bytes_read;
}

size_t libzip_reader_get_remaining(const libzip_reader_t *reader)
{
    size_t remaining = 0;
    
    raise_if_fail(NULL != reader);
    raise_if_fail(NULL != reader->current);
    remaining = reader->remaining;

finally:
    debugf("remaining=%d", remaining);
    return remaining;
}

/* vim: set sw=4 sta et: */
