/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include "libzip-private.h"

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

/* ===== private structures ===== */

struct libzip_writer_t {
    libzip_stream_t *stream;
    libzip_encoder_t *encoder;

    libzip_entry_t *current;
    libzip_entry_t *entries;
    unsigned entry_count;
   
    libzip_compression_t compression;
    libzip_winzip_aes_t winzip_aes;
    char *password;
};

/* ===== libzip_writer ===== */

libzip_writer_t* libzip_writer_create(libzip_stream_t *stream)
{
    libzip_writer_t *writer = NULL;

    raise_if_fail(NULL != stream);
    writer = libzip_new(libzip_writer_t, 1);

    writer->stream = stream;
    writer->compression = LIBZIP_COMPRESSION_STORE;
    
finally:
    return writer;
}

void libzip_writer_set_password(libzip_writer_t *writer, const char* password)
{
    return_if_fail(NULL != writer);

    if (writer->password) {
        free(writer->password);
        writer->password = NULL;
    }

    if (password)
        writer->password = strdup(password);
}

bool libzip_writer_set_compression(libzip_writer_t *writer, libzip_compression_t compression)
{
    return_val_if_fail(NULL != writer, false);
    return_val_if_fail(LIBZIP_COMPRESSION_WINZIP_AES != compression, false);
    return_val_if_fail(LIBZIP_COMPRESSION_STORE == compression ||
                       LIBZIP_COMPRESSION_DEFLATE == compression, false);

    writer->compression = compression;

    return true;
}

bool libzip_writer_set_winzip_aes_mode(libzip_writer_t *writer, libzip_aes_mode_t mode)
{
    return_val_if_fail(NULL != writer, false);
    return_val_if_fail(mode <= LIBZIP_AES_MODE_256, false);

    writer->winzip_aes.mode = mode;
    
    return true;
}

size_t libzip_writer_write_current(libzip_writer_t *writer, const void *buffer, size_t length)
{
    size_t bytes_written = 0;
    
    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->encoder);
    bytes_written = libzip_encoder_write(writer->encoder, buffer, length);

finally:
    return bytes_written;
}

static bool libzip_writer_close_entry(libzip_writer_t *writer, libzip_entry_t *entry)
{
    bool success = true;

    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->stream);
    raise_if_fail(NULL != writer->encoder);
    raise_if_fail(NULL != entry);

    if (!libzip_writer_write_current(writer, NULL, 0))
        raise("Failed to finish data stream");

    if (!libzip_encoder_stat(writer->encoder, &entry->header.uncompressed_size,
                             &entry->header.compressed_size, &entry->header.checksum))
        raise("Failed to read entry stats");

    if (libzip_stream_seek(writer->stream, entry->offset, SEEK_SET)) {
        entry->header.flags&= ~LIBZIP_FLAG_DESCRIPTOR_NEEDED;

        libzip_debug_print_local_header(&entry->header, entry->filename, entry->extra);
        
        if (sizeof(entry->header) != libzip_stream_write(writer->stream, &entry->header, sizeof(entry->header)))
            raise("Cannot update local header");

        if (!libzip_stream_seek(writer->stream, 0, SEEK_END))
            raise("Cannot seek to end of stream");

        success = true;
    } else {
        libzip_local_descriptor_t descriptor;

        descriptor.signature = LIBZIP_SIGNATURE_LOCAL_DESCRIPTOR;
        descriptor.uncompressed_size = entry->header.uncompressed_size;
        descriptor.compressed_size = entry->header.compressed_size;
        descriptor.checksum = entry->header.checksum;
        
        libzip_debug_print_local_descriptor(&descriptor);

        if (sizeof(descriptor) != libzip_stream_write(writer->stream, &descriptor, sizeof(descriptor)))
            raise("Cannot write local descriptor");

        success = true;
    }

finally:
    if (writer->encoder) {
        libzip_encoder_free(writer->encoder);
        writer->encoder = NULL;
    }

    return success;
}

libzip_entry_t* libzip_writer_add_entry(libzip_writer_t *writer, const char *filename, const time_t *timestamp)
{
    libzip_entry_t *entry = NULL;

    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->stream);
    raise_if_fail(NULL != filename);

    entry = libzip_entry_new(filename, timestamp);
    raise_if_fail(NULL != entry);

    if (writer->current) {
        if (false == libzip_writer_close_entry(writer, writer->current))
            raise("Failed to close current entry");

        writer->current = writer->current->next = entry;
    } else {
        writer->current = writer->entries = entry;
    }

    ++writer->entry_count;

    entry->offset = libzip_stream_tell(writer->stream);
    entry->header.compression = writer->compression;

    if (LIBZIP_AES_MODE_NONE != writer->winzip_aes.mode) {
        writer->winzip_aes.header.signature = LIBZIP_EXTRA_SIGNATURE_WINZIP_AES;
        writer->winzip_aes.header.length = (sizeof(writer->winzip_aes) - sizeof(writer->winzip_aes.header));
        writer->winzip_aes.version = LIBZIP_WINZIP_AES_VERSION_AE2;
        writer->winzip_aes.vendor = LIBZIP_WINZIP_AES_VENDOR_AE;
        writer->winzip_aes.compression = entry->header.compression;
        entry->header.compression = LIBZIP_COMPRESSION_WINZIP_AES;
        libzip_entry_add_extra(entry, &writer->winzip_aes.header);
    }

    debugf("writer->compression: %d", writer->compression);
    debugf("entry->header.compression: %d", entry->header.compression);

    raise_if_fail(NULL == writer->encoder);
    writer->encoder = libzip_encoder_new(writer->stream, entry, writer->password);
    raise_if_fail(NULL != writer->encoder);

    libzip_debug_print_local_header(&entry->header, filename, entry->extra);
    libzip_stream_write(writer->stream, &entry->header, sizeof(entry->header));
    libzip_stream_write(writer->stream, filename, strlen(filename));

    if (entry->extra)
        libzip_stream_write(writer->stream, entry->extra, entry->header.extra_length);

finally:
    return entry;
}

libzip_entry_t *libzip_writer_add_file(libzip_writer_t *writer, const char *filename, const char *path)
{
    bool success = false;
    libzip_entry_t *entry = NULL;

    struct stat filestat;
    FILE *istream = NULL;

    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->stream);
    raise_if_fail(NULL != path);

    if (NULL == filename)
        filename = path;

    if (0 == stat(path, &filestat))
        istream = fopen(path, "rb");
    
    if (istream) {
        entry = libzip_writer_add_entry(writer, filename, &filestat.st_mtime);
    } else {
        raise("Cannot open file");
    }

    if (entry) {
        uint8_t buffer[65536];
        size_t length = 0;
        
        while(0 != (length = fread(buffer, 1, sizeof(buffer), istream))) {
            if (false == libzip_writer_write_current(writer, buffer, length))
                raise("Encoding to stream failed");
        }

        if (false == feof(istream))
            raise("Reading from file failed");

        success = true;
    }

finally:
    if (!success)
        entry = NULL;

    if (istream) 
        fclose(istream);
        
    return entry;
}

libzip_entry_t* libzip_writer_add_blob(libzip_writer_t *writer, const char *filename, const time_t *timestamp,
                                       const void *data, size_t length)
{
    bool success = false;
    libzip_entry_t *entry = NULL;

    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->stream);
    raise_if_fail(NULL != filename);
    raise_if_fail(NULL != data);

    if (NULL == (entry = libzip_writer_add_entry(writer, filename, timestamp)))
        raise_if_fail("Failed to create entry in ZIP file");
    if (false == libzip_writer_write_current(writer, data, length))
        raise("Encoding to stream failed");

    success = true;

finally:
    if (!success)
        entry = NULL;

    return entry;
}

libzip_entry_t* libzip_writer_add_text(libzip_writer_t *writer, const char *filename, const time_t *timestamp,
                                       const char *text, int length)
{
    bool success = false;
    libzip_entry_t *entry = NULL;

    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->stream);
    raise_if_fail(NULL != filename);
    raise_if_fail(NULL != text);

    if (-1 == length)
        length = strlen(text);

    if ((entry = libzip_writer_add_entry(writer, filename, timestamp))) {
#ifdef LINE_BREAK_CONVERSION
        const size_t textlen = length;
        const char *curr, *next = text;
        const char *tail = text + textlen;
        
        while(next && next < tail) {
            const char *end = next = strchr(curr = next, '\n');
            
            if (next)
                ++next;

            if (NULL == end)
                end = curr + strlen(curr) - 1;
            while(end > curr && strchr("\r\n\t ", end[-1]))
                --end;

            if (false == libzip_writer_write_current(writer, curr, end - curr))
                raise("Encoding to stream failed");
#ifdef __WIN32      
            if (2 != libzip_writer_write_current(writer, "\r\n", 2))
                raise("Encoding to stream failed");
#else
            if (1 != libzip_writer_write_current(writer, "\n", 1))
                raise("Encoding to stream failed");
#endif
        }

        success = true;
#else
        entry->internal_attributes|= LIBZIP_INTERNAL_ATTRIBUTE_ASCII;
        success = libzip_writer_write_current(writer, text, length);
#endif        
    }

finally:
    if (!success)
        entry = NULL;
    
    return entry;
}

static size_t libzip_writer_write_directory_record(libzip_writer_t *writer, libzip_entry_t *entry)
{
    size_t bytes_written = 0;
    libzip_directory_record_t record;

    raise_if_fail(NULL != writer);
    raise_if_fail(NULL != writer->stream);
    raise_if_fail(NULL != entry);

    record.signature = LIBZIP_SIGNATURE_DIRECTORY_RECORD;
    record.creator_version = LIBZIP_SPECIFICATION_VERSION(2, 0);
    record.filesystem = LIBZIP_FILESYSTEM_MSDOS;
    record.feature_level = entry->header.feature_level;
    record.flags = entry->header.flags;
    record.compression = entry->header.compression;
    record.last_modified = entry->header.last_modified;
    record.checksum = entry->header.checksum;
    record.compressed_size = entry->header.compressed_size;
    record.uncompressed_size = entry->header.uncompressed_size;
    record.filename_length = entry->header.filename_length;
    record.extra_length = entry->header.extra_length;
    record.comment_length = 0;
    record.disk_number_start = 0;
    record.internal_attributes = entry->internal_attributes;
    record.external_attributes = 0x00;
    record.offset_of_local_header = entry->offset;

    libzip_debug_print_directory_record(&record, entry->filename, entry->extra);
    bytes_written+= libzip_stream_write(writer->stream, &record, sizeof(record));
    bytes_written+= libzip_stream_write(writer->stream, entry->filename, record.filename_length);
    
    if (entry->extra)
        bytes_written+= libzip_stream_write(writer->stream, entry->extra, entry->header.extra_length);

finally:
    return bytes_written;
}

static void libzip_writer_write_end_of_directory(libzip_writer_t *writer, size_t offset, size_t length)
{
    libzip_end_of_directory_t end;

    return_if_fail(NULL != writer);
    return_if_fail(NULL != writer->stream);

    end.signature = LIBZIP_SIGNATURE_END_OF_DIRECTORY;
    end.disk_number = 0;
    end.central_directory_disk = 0;
    end.entries_on_disk_count = writer->entry_count;
    end.total_entry_count = writer->entry_count;
    end.length = length;
    end.offset = offset;
    end.comment_length = 0;

    libzip_debug_print_end_of_directory(&end);
    libzip_stream_write(writer->stream, &end, sizeof(end));
}

static void libzip_writer_write_directory(libzip_writer_t *writer)
{
    libzip_entry_t *entry;
    size_t offset, length = 0;
        
    return_if_fail(NULL != writer);
    return_if_fail(NULL != writer->stream);

    offset = libzip_stream_tell(writer->stream);

    for(entry = writer->entries; entry; entry = entry->next)
        length+= libzip_writer_write_directory_record(writer, entry);

    libzip_writer_write_end_of_directory(writer, offset, length);
}

void libzip_writer_close(libzip_writer_t *writer)
{
    if (writer) {
        libzip_entry_t *entry;

        if (writer->current)
            libzip_writer_close_entry(writer, writer->current);
        
        libzip_writer_write_directory(writer);
        libzip_stream_flush(writer->stream);

        while((entry = writer->entries)) {
            writer->entries = entry->next;
            libzip_entry_free(entry);
        }
        
        if (writer->password)
            free(writer->password);

        free(writer);
    }
}

/* vim: set sw=4 sta et: */
