/* A simple library for accessing PKZIP archive files
 * Copyright (C) 2005 Mathias Hasselmann <mathias@taschenorakel.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/**
 * \file libzip.c
 * Library management functions.
 */

#include "libzip-private.h"
#include "fileencrypt/prng.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#ifdef __WIN32
#include <windows.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#endif

/* ===== macro definitions ===== */

#define LIBZIP_RANDOM_DEVICE_DEFAULT "/dev/urandom"

/* ===== global variables ===== */

/** 
 * Level of debug messages which shall be printed.
 * @see debug
 */
int libzip_debug_level = 0;

#ifndef __WIN32

/** 
 * Filename of the random number generator device.
 * @see libzip_random_fd
 */
static char *libzip_random_device = NULL;

/** 
 * File descriptor of the random number generator device.
 * @see random_device
 */
static int libzip_random_fd = -1;

#endif

/**
 * Dr. Gladman's pseudo random number generator.
 */
prng_ctx libzip_prng;

/* ===== debugging ===== */

/** 
 * Modifies debugging status of this library.
 * Default is not to use debugging.
 *
 * @param level New debugging level.
 */
void libzip_set_debug_level(int level)
{
    libzip_debug_level = level;
}

static char libzip_pchar(char ch)
{
    return (isprint(ch) ? ch : '?');
}

static void libzip_debug_print_winzip_aes(libzip_winzip_aes_t* aes)
{
    return_if_fail(NULL != aes);
    return_if_fail(LIBZIP_EXTRA_SIGNATURE_WINZIP_AES == aes->header.signature);
    
    if (libzip_debug_level & 2) {
        printf("[WINZIP-AES]\n");
        printf("- version: %d\n",           aes->version);
        printf("- vendor: %#04x (%c%c)\n",  aes->vendor, 
                                            libzip_pchar(aes->vendor & 255),
                                            libzip_pchar((aes->vendor >> 8) & 255));
        printf("- mode: %d\n",              aes->mode);
        printf("- compression: %d\n",       aes->compression);
    }
}

static void libzip_debug_print_extra(uint8_t *head, uint8_t *tail)
{
    return_if_fail(NULL != head);
    return_if_fail(NULL != tail);
    return_if_fail(head < tail);
    
    if (libzip_debug_level & 2) {
        while(head < tail) {
            libzip_extra_t *extra = (libzip_extra_t*) head;
                    
            printf("[EXTRA-HEADER]\n");
            printf("- signature: %#04x\n",  extra->signature);
            printf("- length: %d\n",        extra->length);

            switch(extra->signature) {
                case LIBZIP_EXTRA_SIGNATURE_WINZIP_AES:
                    libzip_debug_print_winzip_aes((libzip_winzip_aes_t*) extra);
                    break;
            }
            
            head+= extra->length + sizeof(*extra);
        }
    }
}

void libzip_debug_print_local_header(libzip_local_header_t *header, const char *filename, uint8_t *extra)
{
    return_if_fail(NULL != header);
    return_if_fail(LIBZIP_SIGNATURE_LOCAL_HEADER == header->signature);

    if (libzip_debug_level & 2) {
        printf("[LOCAL-HEADER]\n");
        printf("- signature: %#08x\n",      header->signature);
        printf("- feature_level: %02d\n",   header->feature_level);
        printf("- flags: %#04x\n",          header->flags);
        printf("- compression: %#02x\n",    header->compression);
        printf("- last_modified: %#08x\n",  header->last_modified);
        printf("- checksum: %#08x\n",       header->checksum);
        printf("- compressed_size: %d\n",   header->compressed_size);
        printf("- uncompressed_size: %d\n", header->uncompressed_size);
        printf("- filename_length: %d\n",   header->filename_length);
        printf("- extra_length: %d\n",      header->extra_length);

        if (filename)
            printf("- filename: '%s'\n",    filename);
        if (extra)
            libzip_debug_print_extra(extra, extra + header->extra_length);

        return_if_fail(!header->filename_length == !filename);
        return_if_fail(!header->extra_length == !extra);
    }
}

void libzip_debug_print_directory_record(libzip_directory_record_t *record, const char *filename, uint8_t *extra)
{
    return_if_fail(NULL != record);
    return_if_fail(LIBZIP_SIGNATURE_DIRECTORY_RECORD == record->signature);
    
    if (libzip_debug_level & 2) {
        printf("[DIRECTORY-RECORD]\n");
        printf("- signature: %#08x\n",              record->signature);
        printf("- creator_version: %#02x\n",        record->creator_version);
        printf("- filesystem: %#02x\n",             record->filesystem);
        printf("- feature_level: %02d\n",           record->feature_level);
        printf("- flags: %#04x\n",                  record->flags);
        printf("- compression: %#02x\n",            record->compression);
        printf("- last_modified: %#08x\n",          record->last_modified);
        printf("- checksum: %#08x\n",               record->checksum);
        printf("- compressed_size: %d\n",           record->compressed_size);
        printf("- uncompressed_size: %d\n",         record->uncompressed_size);
        printf("- filename_length: %d\n",           record->filename_length);
        printf("- extra_length: %d\n",              record->extra_length);
        printf("- comment_length: %d\n",            record->comment_length);
        printf("- disk_number_start: %d\n",         record->disk_number_start);
        printf("- internal_attributes: %x\n",       record->internal_attributes);
        printf("- external_attributes: %o\n",       record->external_attributes);
        printf("- offset_of_local_header: %d\n",    record->offset_of_local_header);

        if (filename)
            printf("- filename: '%s'\n",        filename);
        if (extra)
            libzip_debug_print_extra(extra, extra + record->extra_length);

        return_if_fail(!record->filename_length == !filename);
        return_if_fail(!record->extra_length == !extra);
    }
}

void libzip_debug_print_end_of_directory(libzip_end_of_directory_t *end)
{
    return_if_fail(NULL != end);
    return_if_fail(LIBZIP_SIGNATURE_END_OF_DIRECTORY == end->signature);
    
    if (libzip_debug_level & 2) {
        printf("[END-OF-DIRECTORY]\n");
        printf("- signature: %#08x\n",              end->signature);
        printf("- disk_number: %d\n",               end->disk_number);
        printf("- central_directory_disk: %d\n",    end->central_directory_disk);
        printf("- entries_on_disk_count: %d\n",     end->entries_on_disk_count);
        printf("- total_entry_count: %d\n",         end->total_entry_count);
        printf("- length: %d\n",                    end->length);
        printf("- offset: %d\n",                    end->offset);
        printf("- comment_length: %d\n",            end->comment_length);
    }
}

void libzip_debug_print_local_descriptor(libzip_local_descriptor_t *descriptor)
{
    return_if_fail(NULL != descriptor);
    return_if_fail(LIBZIP_SIGNATURE_LOCAL_DESCRIPTOR == descriptor->signature);

    if (libzip_debug_level & 2) {
        printf("[LOCAL-DESCRIPTOR]\n");
        printf("- signature: %#08x\n",          descriptor->signature);
        printf("- checksum: %#08x\n",           descriptor->checksum);
        printf("- compressed_size: %d\n",       descriptor->compressed_size);
        printf("- uncompressed_size: %d\n",     descriptor->uncompressed_size);
    }
}

/* ===== pseudo random number generator ===== */

/** 
 * Chooses a device for pseudo random number generation. If never called a 
 * default device is used. On Unix-like operation systems <b>/dev/urandom</b> 
 * is the default device.
 *
 * @param filename Filename of the random number generator device.
 * Pass NULL for using the default device.
 */
void libzip_set_random_device(const char *filename)
{
#ifndef __WIN32
    
    if (libzip_random_device) {
        free(libzip_random_device);
        libzip_random_device = NULL;
    }

    if (-1 != libzip_random_fd) {
        close(libzip_random_fd);
        libzip_random_fd = -1;
    }

    if (filename) {
        libzip_random_device = strdup(filename);
    }

#endif
}

/**
 * Generates random bits for Dr. Gladman's pseudo random number generator.
 */
static int libzip_entropy(uint8_t *buffer, size_t length)
{
    ssize_t bytes_written = -1;

    raise_if_fail(NULL != buffer);

#ifdef __WIN32

    LARGE_INTEGER pentium_tsc;

    if (length > sizeof(pentium_tsc))
        length = sizeof(pentium_tsc);

    QueryPerformanceCounter(&pentium_tsc);
    memcpy(buffer, &pentium_tsc, length);
    bytes_written = length;

#else

    if (-1 == libzip_random_fd) {
        libzip_random_fd = open(
            libzip_random_device ? libzip_random_device :
            LIBZIP_RANDOM_DEVICE_DEFAULT, O_RDONLY);
    }

    raise_if_fail(-1 != libzip_random_fd);
    bytes_written = read(libzip_random_fd, buffer, length);
    raise_if_fail(-1 != bytes_written);

#endif

finally:
    return bytes_written;
}

/**
 * Generates random bits of cryptographic quality.
 */
void libzip_random(uint8_t *buffer, size_t length)
{
    return_if_fail(NULL != buffer);
    prng_rand(buffer, length, &libzip_prng);
}

/* ===== library initialization ===== */

/**
 * Executes initialization code for this library.
 */
static __attribute__((constructor)) void libzip_init()
{
    const char *libzip_debug = getenv("LIBZIP_DEBUG");

    if (libzip_debug)
        libzip_set_debug_level(atoi(libzip_debug));
    
    memset(&libzip_prng, 0, sizeof(libzip_prng));
    prng_init(libzip_entropy, &libzip_prng);
}

/**
 * Executes cleanup code for this library.
 */
static __attribute__((destructor)) void libzip_finish()
{
    prng_end(&libzip_prng);
    libzip_set_random_device(NULL);
}

const char libzip_copyright[] __attribute__((section(".libzip_copyright"))) =
    "libzip is a simple library for accessing PKZIP archive files\n"
    "Copyright (C) 2005 by Mathias Hasselmann <mathias@taschenorakel.de>";

/* vim: set sw=4 sta et: */
