#ifndef LIBZIP_INCLUDED
#define LIBZIP_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>
#include <stdio.h>
#include <time.h>

/**
 * \file libzip.h
 * Public library header.
 */

/* ===== enumerations and constants ===== */

/**
 * Compression modes of the PKZIP file format. 
 */
typedef enum {
    /**
     * Do not compress at all. 
     */
    LIBZIP_COMPRESSION_STORE = 0,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_SHRINK,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_REDUCE_1,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_REDUCE_2,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_REDUCE_3,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_REDUCE_4,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_IMPLODE,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_TOKENIZE,

    /** 
     * Use deflate compression with a window size of 32 KB
     * (<a href="http://www.ietf.org/rfc/rfc1951.txt">RFC&nbsp;1951</a>).
     */
    LIBZIP_COMPRESSION_DEFLATE,
    
    /** 
     * Use deflate compression with a window size of 64 KB
     * (<a href="http://www.ietf.org/rfc/rfc1951.txt">RFC&nbsp;1951</a>).
     */
    LIBZIP_COMPRESSION_DEFLATE64,


    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_PKWARE_IMPLODE,

    /**
     * Not supported.
     */
    LIBZIP_COMPRESSION_PKWARE_RESERVED,

    /**
     * BZip/2 compression.
     * Not supported.
     */
    LIBZIP_COMPRESSION_PKWARE_BZIP2,

    /**
     * WinZIP WavPack compression.
     * Not supported.
     */
    LIBZIP_COMPRESSION_WINZIP_WAVPACK = 97,

    /**
     * WinZIP PPMd compression.
     */
    LIBZIP_COMPRESSION_WINZIP_PPMD = 98,

    /**
     * WinZIP AES encryption. Real compression method is stored in the 
     * libzip_winzip_aes_t header.
     */
    LIBZIP_COMPRESSION_WINZIP_AES = 99
} libzip_compression_t;

/**
 * Key lengths for AES encryption.
 */
typedef enum {
    /**
     * Do not use AES encryption at all. 
     */
    LIBZIP_AES_MODE_NONE =       0x00,
    
    /** 
     * Enable AES encryption and use keys of 128 bit length.
     */
    LIBZIP_AES_MODE_128 =        0x01,
    
    /** 
     * Enable AES encryption and use keys of 256 bit length.
     * Effective key length is 160 bits, as HMAC-SHA1 is used
     * for generating entropy.
     */
    LIBZIP_AES_MODE_192 =        0x02,
    
    /** 
     * Enable AES encryption and use keys of 256 bit length.
     * Effective key length is 160 bits, as HMAC-SHA1 is used
     * for generating entropy.
     */
    LIBZIP_AES_MODE_256 =        0x03
} libzip_aes_mode_t;

/* ===== type definitions ===== */

/** Abstract streaming interface for libzip. */
typedef struct libzip_stream_t libzip_stream_t;

/** Class for reading PKZIP file archives. */
typedef struct libzip_reader_t libzip_reader_t;

/** Class for writing PKZIP file archives. */
typedef struct libzip_writer_t libzip_writer_t;

/** A file entry of a PKZIP file archives. */
typedef struct libzip_entry_t libzip_entry_t;

/* ===== generic functions ===== */

void libzip_set_debug_level(int level);
void libzip_set_random_device(const char *filename);

/* ===== stream handling ===== */

libzip_stream_t* libzip_stream_open(const char *filename, const char *mode);
libzip_stream_t* libzip_stdio_stream_open(FILE *stdio_stream);
bool libzip_stream_close(libzip_stream_t *stream);

/* ===== pkzip file readers ===== */

libzip_reader_t* libzip_reader_open(libzip_stream_t *stream);
void libzip_reader_close(libzip_reader_t *reader);

libzip_entry_t* libzip_reader_find_entry(libzip_reader_t *reader, const char *filename, bool case_sensitive);
libzip_entry_t* libzip_reader_next_entry(libzip_reader_t *reader, libzip_entry_t* entry);

bool libzip_reader_seek(libzip_reader_t *reader, const libzip_entry_t* entry);
size_t libzip_reader_read_current(libzip_reader_t *reader, void *buffer, size_t length);
size_t libzip_reader_get_remaining(const libzip_reader_t *reader);

void libzip_reader_set_password(libzip_reader_t *reader, const char* password);

/* ===== pkzip file writers ===== */

libzip_writer_t* libzip_writer_create(libzip_stream_t *stream);
void libzip_writer_close(libzip_writer_t *writer);

libzip_entry_t* libzip_writer_add_entry(libzip_writer_t *writer, const char *filename, const time_t *timestamp);
libzip_entry_t* libzip_writer_add_file(libzip_writer_t *writer, const char *filename, const char *path);
libzip_entry_t* libzip_writer_add_blob(libzip_writer_t *writer, const char *filename, const time_t *timestamp,
                                       const void *data, size_t length);
libzip_entry_t* libzip_writer_add_text(libzip_writer_t *writer, const char *filename, const time_t *timestamp,
                                       const char *text, int length);
size_t libzip_writer_write_current(libzip_writer_t *writer, const void *buffer, size_t length);

void libzip_writer_set_password(libzip_writer_t *writer, const char* password);
bool libzip_writer_set_compression(libzip_writer_t *writer, libzip_compression_t compression);
bool libzip_writer_set_winzip_aes_mode(libzip_writer_t *writer, libzip_aes_mode_t mode);

/* ===== pkzip file enties ===== */

const char* libzip_entry_get_filename(const libzip_entry_t *entry);
libzip_compression_t libzip_entry_get_compression(const libzip_entry_t *entry);
time_t libzip_entry_get_timestamp(const libzip_entry_t *entry);
size_t libzip_entry_get_length(const libzip_entry_t *entry);

#ifdef __cplusplus
}
#endif

#endif /* vim: set sw=4 sta et: */
